/** ALOE headers
 */
#include <phal_sw_api.h>
#include <swapi_utils.h>

extern struct utils_itf input_itfs[];
extern struct utils_itf output_itfs[];

extern struct utils_param params[];
extern struct utils_stat stats[];


#define DEF_STR_LEN 128
static char str[DEF_STR_LEN];


#define MAX_IN_ITFS    50

int currentInterfaceIdx;

struct initf {
    int fd;
    int rcv_len;
    int long_in;
    struct utils_itf *uitf;
};


struct initf initfs[MAX_IN_ITFS];


#define MAX_OUT_ITFS    50

struct outitf {
    int fd;
    int long_out;
    struct utils_itf *uitf;
};


struct outitf outitfs[MAX_OUT_ITFS];

int nof_input_itfs, nof_output_itfs, nof_stats;

int proc_counter;
int st_inpkts,st_outpkts;

int Init()
{
    int i,n;

    if (!CreateLog()) {
        printf("Error creating log\n");
        return 0;
    }

    if (params[0].name != NULL) {
        n = InitParamFile();
        if (n < 0) {
            printf("Error initiating params file\n");
            return 0;
        }

        for (i = 0; params[i].name; i++) {
            n = GetParam(params[i].name,
                    params[i].value, params[i].type, params[i].size);
            if (!n) {
                printf("Error getting parameter %s\n", params[i].name);
                return 0;
            }
        }
    }

    ConfigInterfaces();

    for (i = 0; input_itfs[i].name && i < MAX_IN_ITFS; i++) {
        initfs[i].fd = CreateItf(input_itfs[i].name, FLOW_READ_ONLY);
        if (initfs[i].fd < 0) {
            printf("Caution input interface %s not connected\n", input_itfs[i].name);
        } 
        initfs[i].rcv_len = 0;
        initfs[i].uitf = &input_itfs[i];
        if (initfs[i].uitf->get_block_sz == NULL) {
            initfs[i].long_in = initfs[i].uitf->max_buffer_len;
        }
    }

    nof_input_itfs = i;
    if (i == MAX_IN_ITFS) {
        printf("Caution some input interfaces may have not been initialized\n");
    }

    for (i = 0; output_itfs[i].name && i < MAX_OUT_ITFS; i++) {
        outitfs[i].fd = CreateItf(output_itfs[i].name, FLOW_WRITE_ONLY);
        if (outitfs[i].fd < 0) {
            printf("Caution output interface %s not connected\n", output_itfs[i].name);
        } 
        outitfs[i].uitf = &output_itfs[i];
    }

    nof_output_itfs = i;
    if (i==MAX_OUT_ITFS) {
        printf("Caution some output interfaces may have not been initialized\n");
    }

    for (i = 0; stats[i].name; i++) {
        if (!stats[i].id) {
            printf("Error configuring stat %s: Error missing id pointer for stats\n", stats[i].name);
            return 0;
        }
        if (stats[i].update_mode != OFF && !stats[i].value) {
            printf("Error configuring stat %s: Automatic mode needs a value pointer\n",stats[i].name);
            return 0;
        }
        *stats[i].id = InitStat(stats[i].name,
                stats[i].type, stats[i].size);
        if (*stats[i].id < 0) {
            printf("Error configuring stat %s: Error initiating with PHAL\n", stats[i].name);
            return 0;
        }
        if (stats[i].value) {
            SetStatsValue(*stats[i].id, stats[i].value, stats[i].size);
        }
    }

    proc_counter = InitCounter("PROC");
    if (proc_counter<0) {
        printf("Error creating proc counter\n");
        return 0;
    }

    st_inpkts = InitStat("INBYTES",STAT_TYPE_INT,1);
    if (st_inpkts<0) {
        printf("Error creating stat inbytes\n");
        return 0;
    }

    st_outpkts = InitStat("OUTBYTES",STAT_TYPE_INT,1);
    if (st_outpkts<0) {
        printf("Error creating stat outbytes\n");
        return 0;
    }

    nof_stats = i;

    return InitCustom();
}

int SendItf(int idx, int len)
{
    int i = idx, n;

    if (idx < 0 || idx > nof_output_itfs) {
        sprintf(str,"Invalid idx %d while sending\n", idx);
        WriteLog(1,str);
        return -1;
    }

    if (!len)
        return 0;

    if (outitfs[i].fd<=0) {
        #ifdef VERBOSE
        printf("Caution: trying to write to unconnected interface %s\n",outitfs[i].uitf->name);
        #endif
        return 1;
    }

    /* write to itf */
    n = WriteItf(outitfs[i].fd, outitfs[i].uitf->buffer, len*outitfs[i].uitf->sample_sz);
    if (n < 0) {
        sprintf(str,"Error writing to itf %s\n", outitfs[i].uitf->name);
        WriteLog(1,str);
        return 0;
    } else if (!n) {
        sprintf(str,"Caution, missing packet due to full buffer for itf %s\n", outitfs[i].uitf->name);
        WriteLog(1,str);
    }

    return n/outitfs[i].uitf->sample_sz;
}

int Run()
{
    int i,n;
    int block_sz;
    int nbytes=0;

    for (i=0; i < nof_stats; i++) {
        if (stats[i].update_mode == READ) {
            if (!stats[i].value) {
                sprintf(str,"Error in stat %s: Mode is READ but any value was defined\n",stats[i].name);
                WriteLog(1,str);
            } else {
                GetStatsValue(*stats[i].id, stats[i].value, stats[i].size);
            }
        }
    }

    for (i = 0; i < nof_input_itfs; i++) {
        currentInterfaceIdx=i;
        if (initfs[i].uitf->get_block_sz != NULL) {
            block_sz = initfs[i].uitf->get_block_sz();
            if (block_sz != initfs[i].long_in) {
                if (block_sz > initfs[i].uitf->max_buffer_len) {
                    sprintf(str, "Error invalid block size for itf %s. Bigger than buffer.\n",initfs[i].uitf->name);
                    WriteLog(1,str);
                    return 0;
                }
                if (block_sz + initfs[i].rcv_len > initfs[i].uitf->max_buffer_len) {
                    sprintf(str, "Error changing min sample length for itf %s: too large\n", initfs[i].uitf->name);
                    WriteLog(1,str);
                    initfs[i].rcv_len = 0;
                }
                initfs[i].long_in = block_sz;
            }
        }
        /* receive a block of data */
        if (initfs[i].fd>0) {
            do {

                n = ReadItf(initfs[i].fd, (void*) (((char*) initfs[i].uitf->buffer) + initfs[i].rcv_len),
                        initfs[i].long_in*initfs[i].uitf->sample_sz - initfs[i].rcv_len);
                if (n < 0) {
                    sprintf(str,"Error reading from itf %s\n",initfs[i].uitf->name);
                    WriteLog(1,str);
                    return 0;
                } else if (!n) {
                    break;
                }
                initfs[i].rcv_len += n;
            } while (n
                && (initfs[i].rcv_len < initfs[i].long_in*initfs[i].uitf->sample_sz
                        || (initfs[i].uitf->get_block_sz==NULL && initfs[i].rcv_len > 0)));
            nbytes += initfs[i].rcv_len;

            /* when enough data is received, process and send it */
            if (initfs[i].rcv_len >= initfs[i].long_in*initfs[i].uitf->sample_sz
                    || ((initfs[i].uitf->get_block_sz==NULL || !initfs[i].uitf->get_block_sz())
                            && initfs[i].rcv_len > 0)) {
                initfs[i].uitf->process_fnc(initfs[i].rcv_len/initfs[i].uitf->sample_sz);
                initfs[i].rcv_len = 0;
            }
        }
    }

    SetStatsValue(st_inpkts,&nbytes,1);
    nbytes=0;

    for (i=0;i<nof_output_itfs;i++) {
        currentInterfaceIdx=i;
        if (outitfs[i].uitf->get_block_sz != NULL
                && outitfs[i].uitf->process_fnc != NULL
                && outitfs[i].fd>0) {

            outitfs[i].long_out = outitfs[i].uitf->get_block_sz();

            if (outitfs[i].long_out > outitfs[i].uitf->max_buffer_len) {
                sprintf(str,"Error requested size exceeds output buffer for itf %s\n", outitfs[i].uitf->name);
                WriteLog(1,str);
            }

            if (outitfs[i].long_out) {
                if (!outitfs[i].uitf->process_fnc(outitfs[i].long_out)) {
                    sprintf(str, "Error generating data for output itf %s.\n",outitfs[i].uitf->name);
                    WriteLog(1,str);
                    return 0;
                }
                SendItf(i,outitfs[i].long_out);
                nbytes+=outitfs[i].long_out;                
            }
        }
    }

    SetStatsValue(st_outpkts,&nbytes,1);

    for (i=0; i < nof_stats; i++) {
        if (stats[i].update_mode == WRITE) {
            if (!stats[i].value) {
                printf("Error in stat %s: Mode is WRITE but any value was defined\n",stats[i].name);
            } else {
                SetStatsValue(*stats[i].id, stats[i].value, stats[i].size);
            }
        }
    }

    return RunCustom();
}

int interval;

void IntervalSet(int numslots)
{
    interval=numslots;
}

/** Main
 */
int main()
{
    interval=1;

    InitPHAL();

    while (1) {
        switch (Status()) {
        case PHAL_STATUS_INIT:
            if (!Init()) {
                printf("Error initiating. Exiting...\n");
                ClosePHAL();
                exit(0);
            }
            break;
        case PHAL_STATUS_RUN:
            StartCounter(proc_counter);
            if (!Run()) {
                printf("Error running. Exiting...\n");
                ClosePHAL();
                exit(0);
            }
            StopCounter(proc_counter);
            break;
        default:
            ClosePHAL();
            exit(0);
            break;
        }
        Relinquish(interval);

    }
    return 1;
}


